#!/usr/bin/perl -w 
###############################################################################
#   afick_common.pl
#      it's a part of the afick project
#
#    Copyright (C) 2002-2004 by Eric Gerbier
#    Bug reports to: gerbier@users.sourceforge.net
#    $Id: afick-common.pl 3146 2013-08-23 07:33:02Z gerbier $
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# common sub and var for afick's tool
###############################################################################
# I use special naming for references :
# $r_name : for a ref to scalar
# $rh_name : for a ref to hashes
# $ra_name : for a ref to array

use strict;
use warnings;

use English qw(-no_match_vars);
use POSIX qw(strftime);
use File::Basename;

use Data::Dumper;

###############################################################################
# where are installed afick modules (.pm)
# will be used to load packages
# and for auto-control
sub get_module_dir() {

	my $module_dir = dirname($PROGRAM_NAME) . '/lib';
	my $linux_dir  = '/usr/lib/afick/lib';
	if ( -e $module_dir ) {

		# for windows and tests
		return $module_dir;
	}
	elsif ( -d $linux_dir ) {

		# for linux
		return $linux_dir;
	}
	else {

		# ?
		return 'lib';
	}
}
##########################################################################
# load libraries
my $module_dir = get_module_dir();
push @INC, $module_dir;

my @modules = (
	'Afick::Constant',
	'Afick::Msg',
	'Afick::Lock',
	'Afick::Tst',
	'Afick::Gen',
	'Afick::Directives',
	'Afick::Macros',
	'Afick::Aliases',

	#'Afick::Cfg',
	#'Afick::Backend',
	#'Afick::Control',
	#'Afick::File',
	#'Afick::Plugins',
	#'Afick::Res',
);

foreach my $module (@modules) {
	## no critic (ProhibitStringyEval,RequireCheckingReturnValueOfEval)
	eval "require $module";
	if ($EVAL_ERROR) {
		die "CRITICAL : can not find $module : $EVAL_ERROR\n";
	}
	else {
		import $module;
	}
}
###############################################################################
# constants
my $EMPTY = Afick::Constant->EMPTY;
my $SPACE = Afick::Constant->SPACE;
my $SLASH = Afick::Constant->SLASH;
###############################################################################
# return default config file according operating system
# and environment variable
sub get_default_config() {
	my $default_config_file;

	if ( exists $ENV{AFICK_CONFIG} ) {
		$default_config_file = $ENV{AFICK_CONFIG};
	}
	elsif ( is_microsoft() ) {

		# config in install directory
		$default_config_file = 'windows.conf';
	}

	# unix/linux
	elsif ( -r '/etc/afick.conf' ) {

		# system
		$default_config_file = '/etc/afick.conf';
	}
	else {

		# config in install directory
		$default_config_file = 'afick.conf';
	}

	return ( -r $default_config_file ) ? $default_config_file : $EMPTY;
}
###############################################################################
# read config file and have some cleaning
sub read_config($$) {
	my $configfile = shift @_;    # config file name
	my $ra_config  = shift @_;    # array to store config lines

	if ( !$configfile ) {
		Afick::Msg->set_error('(read_config) no config file');
		return;
	}
	elsif ( open my $fh_config, '<', $configfile ) {
		while ( my $ligne = <$fh_config> ) {

			# normalize config line
			chomp $ligne;
			remove_trailing_spaces( \$ligne );
			push @{$ra_config}, $ligne;
		}
		close $fh_config
		  or Afick::Msg->set_error(
			"(read_config) can not close $configfile : $ERRNO");
		return 1;
	}
	else {
		Afick::Msg->set_error(
			"(read_config) can not open $configfile : $ERRNO");
		return;
	}
}
###############################################################################
# write to a config file
sub write_config($$) {
	my $configfile = shift @_;    # config file name
	my $ra_config  = shift @_;    # array containing config lines

	if ( open my $fh_config, '>', $configfile ) {
		foreach my $ligne ( @{$ra_config} ) {
			print {$fh_config} $ligne . "\n";
		}
		close $fh_config
		  or Afick::Msg->set_error("can not close $configfile : $ERRNO");
		return 1;
	}
	else {
		Afick::Msg->set_error("can not write to $configfile : $ERRNO");
		return;
	}
}
##########################################################################
# check duplicate for alias, macros
sub check_duplicate($$) {
	my $key     = shift @_;
	my $rh_find = shift @_;

	Afick::Msg->debug("(check_duplicate) $key");

	if ( $key =~ m/^exclude/ ) {

		# allow multi_lines
		# rh_find is to true on first call
		# and then set to false in overload sub
		$rh_find->{$key} = 1 unless exists $rh_find->{$key};
		return 0;
	}
	elsif ( exists $rh_find->{$key} ) {
		Afick::Msg->set_error('find duplicate key');
		return 1;
	}
	else {
		$rh_find->{$key}++;
		return 0;
	}
}
#######################################################
# expand_line
# expand config line : treat jokers, and split into <file attribute flagdir>
sub trait_line ($$$) {
	my $name        = shift @_;    # file/directory name
	my $attribute   = shift @_;    # Myrule, p+u+g
	my $ignore_case = shift @_;

	# if ignore_case change all names
	$name = lc $name if ($ignore_case);

	# for compatibility with old syntax
	$attribute = 'all' if ( !$attribute );

	# Afick::Msg->debug("$name -> $attribute", 1);

	my @tab;
	my $flagdir;                   # true if is a directory

	# expand jokers if exists
	if ( $name =~ m/[*?]/ ) {
		@tab = glob $name;
		Afick::Msg->debug( "(trait_line) expand $name in @tab", 2 );
		if ($ignore_case) {
			foreach (@tab) {
				$_ = lc;
			}
		}
	}

	# test for root
	elsif ( is_fsroot($name) ) {
		$flagdir = 1;
		push @tab, $name;
	}
	elsif ( $name =~ m{/$} ) {

		# remove trailing /
		$name =~ s{/$}{};
		$flagdir = 1;
		push @tab, $name;
	}
	else {
		$flagdir = 0;
		push @tab, $name;
	}
	push @tab, $attribute, $flagdir;
	return @tab;
}
##########################################################################
# a clone from read_configuration (afick.pl)
# but using parameter instead global var
# pb : duplicated code
# TODO : merge the 2 codes (Afick::Cfg)
## no critic (ProhibitManyArgs,ProhibitExcessComplexity)
sub get_configuration($$$$$$) {
	my $configfile   = shift @_;
	my $rh_macro     = shift @_;
	my $rh_alias     = shift @_;
	my $rh_directive = shift @_;
	my $rh_rule      = shift @_;
	my $rh_onlydir   = shift @_;

	my @config;
	if ( !read_config( $configfile, \@config ) ) {

		# warn or warning ?
		Afick::Msg->warning( Afick::Msg->get_error() );
		return 0;
	}

	# some hash to detect multiple lines
	my %h_m_macro;
	my %h_m_directive;
	my %h_m_alias;

	# objects
	my $directives = Afick::Directives->new();
	my $macros     = Afick::Macros->new();
	my $aliases    = Afick::Aliases->new();

	# control chroot
	chroot_check();

	my $nb_pbs  = 0;
	my $line_id = 0;
  LOOP: foreach my $line (@config) {
		$line_id++;

		# clean line
		chomp $line;
		remove_trailing_spaces( \$line );
		next LOOP unless length $line;    # skip blank lines
		next LOOP if ( $line =~ m/^#/ );  # skip comments

		# replace \ by / (for windows)
		my $linerule = reg_name($line);

		# remove trailing slash
		$linerule =~ s{/$}{};

		# environment variables check
		if ( $linerule =~ m/(.*)\${(\w+)}(.*)/ ) {

			if ( exists $ENV{$2} ) {

				# just replace on fly
				my $expand = $ENV{$2};
				$linerule = $1 . $expand . $3;
				Afick::Msg->debug(
"expand $2 into $expand in config file $line (line $line_id)",
					2
				);
			}
			else {

				# problem : env does not exists
				# just skip line
				Afick::Msg->warning(
"(get_config) skipped config file $line (line $line_id) : found unknown environment variable $2"
				);
				next LOOP;
			}
		}

		#Afick::Msg->debug( "config line = $line", 1);

		my @ret;
		## no critic (ProhibitCascadingIfElse)
		if ( @ret = is_macro_line($line) ) {

			# macros lines
			###################
			my $key = shift @ret;
			my $val = shift @ret;

			if (   ( !defined $macros->check_macro( $key, $val, 0 ) )
				or ( check_duplicate( $key, \%h_m_macro ) ) )
			{
				$nb_pbs++;
				next;
			}
			Afick::Msg->debug( "(get_config) found macro $key : $val", 2 );

			# add to Macro for print_config
			$macros->set_macro( $key, $val );
			$h_m_macro{$key} = $val;    # for check_duplicate
		}
		elsif ( @ret = is_directive_line($line) ) {

			# a directive line
			######################
			my $key = shift @ret;
			my $val = shift @ret;
			remove_trailing_spaces( \$val );

			my $ret2 = $directives->check_directive( $key, $val, 0 );
			if (   ( !defined $ret2 )
				or ( check_duplicate( $key, \%h_m_directive ) ) )
			{
				$nb_pbs++;
				next;
			}
			Afick::Msg->debug( "(get_config) found directive $key : $val", 2 );
			$directives->set_directive( $key, $val );
			$h_m_directive{$key} = $val;    # for check_duplicate
		}
		elsif ( @ret = is_alias_line($line) ) {

			# is an alias definition
			########################
			my $key = shift @ret;
			my $val = shift @ret;
			remove_all_spaces( \$val );

			my $decoded = $aliases->check_alias($val);
			if (   ( !defined $decoded )
				or ( check_duplicate( $key, \%h_m_alias ) ) )
			{
				$nb_pbs++;
				next;
			}

			Afick::Msg->debug( "(get_config) alias $key -> $decoded", 2 );
			$aliases->set_alias( $key, $val );
			$h_m_alias{$key} = $val;    # for check_duplicate
		}
		elsif ( @ret = is_negsel_line($linerule) ) {
			my @tab =
			  trait_line( shift @ret, shift @ret, $rh_macro->{'ignore_case'} );
			my $flagdir = pop @tab;     # and ignore
			my $masq    = pop @tab;     # and ignore
			foreach my $elem (@tab) {
				my $chroot_elem = add_chroot($elem);

				$rh_rule->{$chroot_elem} = 0;
				Afick::Msg->debug( "(get_config) negsel $chroot_elem", 2 );
			}
		}
		elsif ( @ret = is_equalsel_line($linerule) ) {
			my @tab =
			  trait_line( shift @ret, shift @ret, $rh_macro->{'ignore_case'} );
			my $flagdir = pop @tab;     # ignore
			my $masqb   = pop @tab;     # brut alias line

			my $masq = $aliases->check_alias($masqb);
			if ( !defined $masq ) {
				$nb_pbs++;
				next;
			}
			foreach my $elem (@tab) {
				my $chroot_elem = add_chroot($elem);
				$rh_rule->{$chroot_elem}    = $masqb;
				$rh_onlydir->{$chroot_elem} = 1;

				Afick::Msg->debug( "(get_config) equalsel $chroot_elem $masq",
					2 );
			}
		}
		elsif ( @ret = is_sel_line($linerule) ) {
			my @tab =
			  trait_line( shift @ret, shift @ret, $rh_macro->{'ignore_case'} );
			my $flagdir = pop @tab;    # ignore
			my $masqb   = pop @tab;    # brut alias line

			my $masq = $aliases->check_alias($masqb);
			if ( !defined $masq ) {
				$nb_pbs++;
				next;
			}
			foreach my $elem (@tab) {
				my $chroot_elem = add_chroot($elem);
				$rh_rule->{$chroot_elem} = $masqb;

				Afick::Msg->debug( "(get_config) sel $chroot_elem $masq", 2 );
			}
		}
		else {
			Afick::Msg->warning(
"(get_config) skipped config file $line (line $line_id) : unknown type"
			);
		}
		## use critic
	}    # foreach my $line

	%{$rh_macro}     = $macros->macros();
	%{$rh_directive} = $directives->directives();
	%{$rh_alias}     = $aliases->aliases();

	# callers do not test code return for now
	return 1;
}
## use critic
##########################################################################
# afick current version : to be common for afick and afick-tk
sub get_afick_version() {
	return '3.4';
}
##########################################################################

1;
